import inspect
import weakref
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

import torch
from torch._subclasses.fake_tensor import FakeCopyMode, FakeTensorMode
from torch.fx.experimental.symbolic_shapes import ShapeEnv

from nnsight.tracing.Node import Node

from .. import util
from ..contexts.Conditional import ConditionalManager
from .util import validate

if TYPE_CHECKING:
    from ..contexts.backends.LocalBackend import LocalMixin
    from ..contexts.Conditional import Conditional
    from ..intervention import InterventionProxy
    from .Bridge import Bridge
    from .Graph import Graph
    from .Node import Node


class Protocol:
    """A `Protocol` represents some complex action a user might want to create a `Node` and `Proxy` for as well as add to a `Graph`.
    Unlike normal `Node` target execution, these have access to the `Node` itself and therefore the `Graph`, enabling more powerful functionality than with just functions and methods.
    """

    redirect: bool = True
    condition: bool = True

    @classmethod
    def add(cls, *args, **kwargs) -> "InterventionProxy":
        """Class method to be implemented in order to add a Node of this Protocol to a Graph."""

        raise NotImplementedError()

    @classmethod
    def execute(cls, node: "Node"):
        """Class method to be implemented which contains the actual execution logic of the Protocol. By default, does nothing

        Args:
            node (Node): Node to execute using this Protocols execution logic.
        """
        pass

    @classmethod
    def compile(cls, node: "Node") -> None:
        pass

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "black", "shape": "ellipse"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class ApplyModuleProtocol(Protocol):
    """Protocol that references some root model, and calls its .forward() method given some input.
    Using .forward() vs .__call__() means it wont trigger hooks.
    Uses an attachment to the Graph to store the model.
    """

    attachment_name = "nnsight_root_module"

    @classmethod
    def add(
        cls, graph: "Graph", module_path: str, *args, hook=False, **kwargs
    ) -> "InterventionProxy":
        """Creates and adds an ApplyModuleProtocol to the Graph.
        Assumes the attachment has already been added via ApplyModuleProtocol.set_module().

        Args:
            graph (Graph): Graph to add the Protocol to.
            module_path (str): Module path (model.module1.module2 etc), of module to apply from the root module.

        Returns:
            InterventionProxy: ApplyModule Proxy.
        """

        value = inspect._empty

        # If the Graph is validating, we need to compute the proxy_value for this node.
        if graph.validate:

            from .Node import Node

            # If the module has parameters, get its device to move input tensors to.
            module: torch.nn.Module = util.fetch_attr(
                cls.get_module(graph), module_path
            )

            try:
                device = next(module.parameters()).device
            except:
                device = None

            # Enter FakeMode for proxy_value computing.
            value = validate(module.forward, *args, **kwargs)

        kwargs["hook"] = hook

        # Create and attach Node.
        return graph.create(
            target=cls,
            proxy_value=value,
            args=[module_path] + list(args),
            kwargs=kwargs,
        )

    @classmethod
    def execute(cls, node: "Node") -> None:
        """Executes the ApplyModuleProtocol on Node.

        Args:
            node (Node): ApplyModule Node.
        """

        module: torch.nn.Module = util.fetch_attr(
            cls.get_module(node.graph), node.args[0]
        )

        try:
            device = next(module.parameters()).device
        except:
            device = None

        args, kwargs = node.prepare_inputs(
            (node.args, node.kwargs), device=device
        )

        module_path, *args = args

        hook = kwargs.pop("hook")

        if hook:
            output = module(*args, **kwargs)
        else:
            output = module.forward(*args, **kwargs)

        node.set_value(output)

    @classmethod
    def set_module(cls, graph: "Graph", module: torch.nn.Module) -> None:
        """Sets the nnsight root module as an attachment on a Graph.

        Args:
            graph (Graph): Graph.
            module (torch.nn.Module): Root module.
        """

        graph.attachments[cls.attachment_name] = module

    @classmethod
    def get_module(cls, graph: "Graph") -> torch.nn.Module:
        """Returns the nnsight root module from an attachment on a Graph.

        Args:
            graph (Graph): Graph

        Returns:
            torch.nn.Module: Root Module.
        """

        return graph.attachments[cls.attachment_name]

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {
                "color": "blue",
                "shape": "polygon",
                "sides": 6,
            },  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class LockProtocol(Protocol):
    """Simple Protocol who's .execute() method does nothing. This means not calling .set_value() on the Node, therefore  the Node won't be destroyed."""

    redirect: bool = False

    @classmethod
    def add(cls, node: "Node") -> "InterventionProxy":

        return node.create(
            proxy_value=None,
            target=cls,
            args=[node],
        )

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "brown", "shape": "ellipse"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument lable key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class GradProtocol(Protocol):
    """Protocol which adds a backwards hook via .register_hook() to a Tensor. The hook injects the gradients into the node's value on hook execution.
    Nodes created via this protocol are relative to the next time .backward() was called during tracing allowing separate .grads to reference separate backwards passes:

    .. code-block:: python
        with model.trace(...):

            grad1 = model.module.output.grad.save()

            model.output.sum().backward(retain_graph=True)

            grad2 = model.module.output.grad.save()

            model.output.sum().backward()

    Uses an attachment to store number of times .backward() has been called during tracing so a given .grad hook is only value injected at the appropriate backwards pass.
    """

    attachment_name = "nnsight_backward_idx"

    @classmethod
    def add(cls, node: "Node") -> "InterventionProxy":

        # Get number of times .backward() was called during tracing from an attachment. Use as Node argument.
        backward_idx = node.graph.attachments.get(cls.attachment_name, 0)

        return node.create(
            proxy_value=node.proxy_value,
            target=cls,
            args=[node, backward_idx],
        )

    @classmethod
    def execute(cls, node: "Node") -> None:

        args, kwargs = node.prepare_inputs((node.args, node.kwargs))

        # First arg is the Tensor to add hook to.
        tensor: torch.Tensor = args[0]
        # Second is which backward pass this Node refers to.
        backward_idx: int = args[1]

        # Hook to remove when hook is executed at the appropriate backward pass.
        hook = None

        def grad(value):

            nonlocal backward_idx

            # If backward_idx == 0, this is the correct backward pass and we should actually execute.
            if backward_idx == 0:

                # Set the value of the Node.
                node.set_value(value)

                if node.attached():

                    # There may be a swap Protocol executed during the resolution of this part of the graph.
                    # If so get it and replace value with it.
                    value = SwapProtocol.get_swap(node.graph, value)

                # Don't execute this hook again.
                backward_idx = -1

                # Remove hook (if this is not done memory issues occur)
                hook.remove()

                return value

            # Otherwise decrement backward_idx
            else:

                backward_idx -= 1

                return None

        # Register hook.
        hook = tensor.register_hook(grad)

    @classmethod
    def increment(cls, graph: "Graph"):
        """Increments the backward_idx attachment to track the number of times .backward() is called in tracing for this Graph.

        Args:
            graph (Graph): Graph.
        """

        backward_idx = graph.attachments.get(cls.attachment_name, 0)

        graph.attachments[cls.attachment_name] = backward_idx + 1

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "green4", "shape": "box"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class SwapProtocol(Protocol):
    """Protocol which adds an attachment to the Graph which can store some value. Used to replace ('swap') a value with another value."""

    attachment_name = "nnsight_swap"

    @classmethod
    def add(cls, node: "Node", value: Any) -> "InterventionProxy":

        return node.create(target=cls, args=[node, value], proxy_value=True)

    @classmethod
    def execute(cls, node: "Node") -> None:

        # In case there is already a swap, get it from attachments.
        swap: "Node" = node.graph.attachments.get(cls.attachment_name, None)

        # And set it to False to destroy it.
        if swap is not None:
            swap.set_value(False)

        # Set the swap to this Node.
        node.graph.attachments[cls.attachment_name] = node

    @classmethod
    def get_swap(cls, graph: "Graph", value: Any) -> Any:
        """Checks if a swap exists on a Graph. If so get and return it, otherwise return the given value.

        Args:
            graph (Graph): Graph
            value (Any): Default value.

        Returns:
            Any: Default value or swap value.
        """

        # Tries to get the swap.
        swap: "Node" = graph.attachments.get(cls.attachment_name, None)

        # If there was one:
        if swap is not None:

            device = None

            def _device(value: torch.Tensor):
                nonlocal device

                device = value.device

            # Get device of default value.
            util.apply(value, _device, torch.Tensor)

            # Get swap Node's value.
            value = util.apply(swap.args[1], lambda x: x.value, type(swap))

            if device is not None:

                def _to(value: torch.Tensor):
                    return value.to(device)

                # Move swap values to default value's device.
                value = util.apply(value, _to, torch.Tensor)

            # Set value of 'swap' node so it destroys itself and listeners.
            swap.set_value(True)

            # Un-set swap.
            graph.attachments[cls.attachment_name] = None

        return value

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "green4", "shape": "ellipse"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge key word


class BridgeProtocol(Protocol):
    """Protocol to connect two Graphs by grabbing a value from one and injecting it into another.
    Uses an attachment to store a Bridge object which references all relevant Graphs and their ordering.
    """

    attachment_name = "nnsight_bridge"
    condition: bool = False

    class BridgeException(Exception):
        def __init__(self):
            super.__init__(
                "Must define a Session context to make use of the Bridge"
            )

    @classmethod
    def add(cls, node: "Node") -> "InterventionProxy":

        bridge = cls.get_bridge(node.graph)
        curr_graph = bridge.peek_graph()
        bridge_proxy = bridge.get_bridge_proxy(
            node, curr_graph.id
        )  # a bridged node has a unique bridge node proxy per graph reference

        # if the bridge node does not exist, create one
        if bridge_proxy is None:
            # Adds a Lock Node. One, so the value from_node isn't destroyed until the to_nodes are done with it,
            # and two acts as an easy reference to the from_node to get its value from the lock Node args.
            lock_node = LockProtocol.add(node).node

            # Args for a Bridge Node are the id of the Graph and node name of the Lock Node.
            bridge_proxy = node.create(
                target=cls,
                proxy_value=node.proxy_value,
                args=[node.graph.id, lock_node.name],
            )
            bridge.add_bridge_proxy(node, bridge_proxy)

        return bridge_proxy

    @classmethod
    def execute(cls, node: "Node") -> None:

        # Gets Bridge object from the Node's Graph.
        bridge = cls.get_bridge(node.graph)

        # Args are Graph's id and name of the Lock Node on it.
        from_graph_id, lock_node_name = node.args

        # Gets the from_node's Graph via its id with the Bridge and get the Lock Node.
        lock_node = bridge.get_graph(from_graph_id).nodes[lock_node_name]

        # Value node is Lock Node's only arg
        value_node: "Node" = lock_node.args[0]
        
        if value_node.done():

            # Set value to that of the value Node.
            node.set_value(value_node.value)

        # Bridge.release tells this Protocol when to release all Lock Nodes as we no longer need the data (useful when running a Graph in a loop, only release on last iteration)
        if bridge.release:

            lock_node.set_value(None)

    @classmethod
    def set_bridge(cls, graph: "Graph", bridge: "Bridge") -> None:
        """Sets Bridge object as an attachment on a Graph.

        Args:
            graph (Graph): Graph.
            bridge (Bridge): Bridge.
        """

        graph.attachments[cls.attachment_name] = weakref.proxy(bridge)

    @classmethod
    def get_bridge(cls, graph: "Graph") -> "Bridge":
        """Gets Bridge object from a Graph. Assumes Bridge has been set as an attachment on this Graph via BridgeProtocol.set_bridge().

        Args:
            graph (Graph): Graph.

        Returns:
            Bridge: Bridge.
        """

        if not cls.has_bridge(graph):
            raise cls.BridgeException()

        return graph.attachments[cls.attachment_name]

    @classmethod
    def has_bridge(cls, graph: "Graph") -> bool:
        """Checks to see if a Bridge was added as an attachment on this Graph via BridgeProtocol.set_bridge().

        Args:
            graph (Graph): Graph

        Returns:
            bool: If Graph has Bridge attachment.
        """

        return cls.attachment_name in graph.attachments

    @classmethod
    def peek_graph(cls, graph: "Graph") -> "Graph":
        """Returns current Intervention Graph.

        Args:
            - graph (Graph): Graph.

        Returns:
            Graph: Graph.
        """

        if not BridgeProtocol.has_bridge(graph):
            return graph
        else:
            bridge = BridgeProtocol.get_bridge(graph)
            return bridge.peek_graph()

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "brown", "shape": "box"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {
                    "color": "gray",
                    "shape": "box",
                },  # Non-node argument display
                {0: {"color": "gray", "shape": "box", "style": "dashed"}},
            ),
            "arg_kname": defaultdict(
                lambda: None, {0: "graph_id"}
            ),  # Arugment label key word
            "edge": defaultdict(lambda: "solid", {0: "dashed"}),
        }  # Argument edge display


class EarlyStopProtocol(Protocol):
    """Protocol to stop the execution of a model early."""

    class EarlyStopException(Exception):
        pass

    @classmethod
    def add(
        cls, graph: "Graph", stop_point_node: Optional["Node"] = None
    ) -> "InterventionProxy":
        return graph.create(
            target=cls,
            proxy_value=None,
            args=([stop_point_node] if stop_point_node is not None else []),
        )

    @classmethod
    def execute(cls, node: "Node") -> None:

        node.set_value(True)

        raise cls.EarlyStopException()

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {
                "color": "red",
                "shape": "polygon",
                "sides": 6,
            },  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class LocalBackendExecuteProtocol(Protocol):

    @classmethod
    def add(cls, object: "LocalMixin", graph: "Graph") -> "InterventionProxy":

        return graph.create(target=cls, proxy_value=None, args=[object])

    @classmethod
    def execute(cls, node: Node) -> None:

        object: "LocalMixin" = node.args[0]

        object.local_backend_execute()

        node.set_value(None)

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {
                "color": "purple",
                "shape": "polygon",
                "sides": 6,
            },  # Node display
            "label": "ExecuteProtocol",
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class ValueProtocol(Protocol):

    @classmethod
    def add(cls, graph: "Graph", default: Any = None) -> "InterventionProxy":

        return graph.create(target=cls, proxy_value=default, args=[default])

    @classmethod
    def execute(cls, node: Node) -> None:
        
        node.set_value(node.args[0])

    @classmethod
    def set(cls, node: Node, value: Any) -> None:

        node.args[0] = value

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "blue", "shape": "box"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class ConditionalProtocol(Protocol):
    """Protocol operating as a conditional statement.
    Uses the ConditionalManager attachment to handle all visited Conditional contexts within a single Intervention Graph.
    Evaluates the condition value of the Conditional as a boolean.

    Example:

        Setup:
            .. code-block:: python
                import torch
                from collections import OrderedDict

                input_size = 5
                hidden_dims = 10
                output_size = 2

                model = nn.Sequential(OrderedDict([
                    ('layer1', torch.nn.Linear(input_size, hidden_dims)),
                    ('layer2', torch.nn.Linear(hidden_dims, output_size)),
                ]))

                input = torch.rand((1, input_size))    å

        Ex 1: The .save() on the model output will only be executed if the condition (x > 0) is evaluated to True.

        .. code-block:: python
            with model.trace(input) as tracer:
                num = 5
                with tracer.cond(x > 0):
                    out = model.output.save()

        Ex 2: The condition is a tensor boolean operation on the Envoy's output InterventionProxy.

        .. code-block:: python
            with model.trace(input) as tracer:
                l1_out = model.layer1.output
                with tracer.cond(l1_out[:, 0] > 0):
                    out = model.output.save()
    """

    attachment_name = "nnsight_conditional_manager"

    @classmethod
    def add(
        cls, graph: "Graph", condition: Union["Node", Any]
    ) -> "InterventionProxy":

        return graph.create(target=cls, proxy_value=True, args=[condition])

    @classmethod
    def execute(cls, node: "Node") -> None:
        """Evaluate the node condition to a boolean.

        Args:
            node (Node): ConditionalProtocol node.
        """

        cond_value = Node.prepare_inputs(node.args[0])
        if cond_value:
            # cond_value is True
            node.set_value(True)
            return

        def update_conditioned_nodes(conditioned_node: "Node") -> None:
            """Recursively decrement the remaining listeners count of all the dependencies of conditioned nodes.

            Args:
                - conditioned_node (Node): Conditioned Node
            """
            for listener in conditioned_node.listeners:
                for listener_arg in listener.arg_dependencies:
                    listener_arg.remaining_listeners -= 1
                    if listener_arg.done() and listener_arg.redundant():
                        listener_arg.destroy()
                update_conditioned_nodes(listener)

        # If the condition value is ignore or evaluated to False, update conditioned nodes
        update_conditioned_nodes(node)

    @classmethod
    def has_conditional(cls, graph: "Graph") -> bool:
        """Checks if the Intervention Graph has a ConditionalManager attached to it.

        Args:
            graph (Graph): Intervention Graph.

        Returns:
            bool: If graph has a ConditionalManager attachement.
        """
        return cls.attachment_name in graph.attachments.keys()

    @classmethod
    def get_conditional(
        cls, graph: "Graph", cond_node_name: str
    ) -> "Conditional":
        """Gets the ConditionalProtocol node by its name.

        Args:
            graph (Graph): Intervention Graph.
            cond_node_name (str): ConditionalProtocol name.

        Returns:
            Node: ConditionalProtocol Node.
        """
        return graph.attachments[cls.attachment_name].get(cond_node_name)

    @classmethod
    def push_conditional(cls, node: "Node") -> None:
        """Attaches a Conditional context to its graph.

        Args:
            node (Node): ConditionalProtocol of the current Conditional context.
        """

        # All ConditionalProtocols associated with a graph are stored and managed by the ConditionalManager.
        # Create a ConditionalManager attachement to the graph if this the first Conditional context to be entered.
        if cls.attachment_name not in node.graph.attachments.keys():
            node.graph.attachments[cls.attachment_name] = ConditionalManager()

        # Push the ConditionalProtocol node to the ConditionalManager
        node.graph.attachments[cls.attachment_name].push(node)

    @classmethod
    def pop_conditional(cls, graph: "Graph") -> None:
        """Pops latest ConditionalProtocol from the ConditionalManager attached to the graph.

        Args:
            graph (Graph): Intervention Graph.
        """
        graph.attachments[cls.attachment_name].pop()

    @classmethod
    def peek_conditional(cls, graph: "Graph") -> "Node":
        """Gets the ConditionalProtocol node of the current Conditional context.

        Args:
            - graph (Graph): Graph.

        Returns:
            Node: ConditionalProtocol of the current Conditional context.
        """
        return graph.attachments[cls.attachment_name].peek()

    @classmethod
    def add_conditioned_node(cls, node: "Node") -> None:
        """Adds a conditioned Node the ConditionalManager attached to its graph.

        Args:
            - node (Node): Conditioned Node.
        """

        node.graph.attachments[cls.attachment_name].add_conditioned_node(node)

    @classmethod
    def is_node_conditioned(cls, node: "Node") -> bool:
        """Checks if the Node is conditoned by the current Conditional context.

        Args:
            - node (Node): Conditioned Node.

        Returns:
            bool: Whether the Node is conditioned.
        """

        return node.graph.attachments[cls.attachment_name].is_node_conditioned(
            node
        )

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {
                "color": "#FF8C00",
                "shape": "polygon",
                "sides": 6,
            },  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display


class UpdateProtocol(Protocol):
    """Protocol to update the value of an InterventionProxy node.

    .. codeb-block:: python
        with model.trace(input) as tracer:
            num = tracer.apply(int, 0)
            num.update(5)
    """

    @classmethod
    def add(
        cls, node: "Node", new_value: Union[Node, Any]
    ) -> "InterventionProxy":
        """Creates an UpdateProtocol node.

        Args:
            node (Node): Original node.
            new_value (Union[Node, Any]): The update value.

        Returns:
            InterventionProxy: proxy.
        """

        return node.create(
            target=cls,
            proxy_value=node.proxy_value,
            args=[
                node,
                new_value,
            ],
        )

    @classmethod
    def execute(cls, node: "Node") -> None:
        """Sets the value of the original node to the new value.
            If the original is defined outside the context, it uses the bridge to get the node.

        Args:
            node (Node): UpdateProtocol node.
        """

        value_node, new_value = node.args
        new_value = Node.prepare_inputs(new_value)

        if value_node.target == BridgeProtocol:
            value_node._value = new_value
            bridge = BridgeProtocol.get_bridge(value_node.graph)
            lock_node = bridge.id_to_graph[value_node.args[0]].nodes[
                value_node.args[1]
            ]
            value_node = lock_node.args[0]

        value_node._value = new_value

        node.set_value(new_value)

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "blue", "shape": "ellipse"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument
            "arg_kname": defaultdict(lambda: None),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument edge display



from __future__ import annotations

from typing import Any, Dict, List, Tuple, Union

import torch
from transformer_lens import HookedTransformer, HookedTransformerConfig
from transformers import BatchEncoding

from .LanguageModel import LanguageModel


class UnifiedTransformer(LanguageModel):
    """UnifiedTransformer is an nnsight wrapper around TransformerLens's HookedTransformer.

    Inputs can be in the form of:
        Prompt: (str)
        Prompts: (List[str])
        Batched prompts: (List[List[str]])
        Tokenized prompt: (Union[List[int], torch.Tensor])
        Tokenized prompts: (Union[List[List[int]], torch.Tensor])
        Direct input: (Dict[str,Any])

    TransformerLens processing arguments can be passed as kwargs to the constructor.
    Pass `processing=False` to call `from_pretrained_no_processing` instead of `from_pretrained`.

    Calls to generate pass arguments downstream to :func:`GenerationMixin.generate`

    """

    def __init__(
        self, model: str, *args, processing: bool = True, **kwargs
    ) -> None:
        if processing:
            hooked_model = HookedTransformer.from_pretrained(model, *args, **kwargs)
        else:
            hooked_model = HookedTransformer.from_pretrained_no_processing(
                model, *args, **kwargs
            )

        self.tokenizer = hooked_model.tokenizer
        self._model: HookedTransformer = None

        super().__init__(hooked_model, tokenizer=self.tokenizer, *args, **kwargs)

    def _prepare_inputs(
        self,
        inputs: Union[
            str,
            List[str],
            List[List[str]],
            List[int],
            List[List[int]],
            torch.Tensor,
            Dict[str, Any],
            BatchEncoding,
        ],
        **kwargs,
    ) -> Tuple[BatchEncoding, int]:
        if isinstance(inputs, dict):

            new_inputs = dict()

            tokenized_inputs = self._tokenize(inputs["input"], **kwargs)

            new_inputs["input"] = tokenized_inputs["input_ids"]

            if "attention_mask" in inputs:
                for ai, attn_mask in enumerate(inputs["attention_mask"]):
                    tokenized_inputs["attention_mask"][
                        ai, -len(attn_mask) :
                    ] = attn_mask

                new_inputs["attention_mask"] = tokenized_inputs["attention_mask"]

            return (BatchEncoding(new_inputs),), len(new_inputs["input"])

        inputs = self._tokenize(inputs, **kwargs)

        if "input_ids" in inputs:
            inputs["input"] = inputs.pop("input_ids")

        return (inputs,), len(inputs["input"])

    def _batch_inputs(
        self,
        batched_inputs: Dict,
        prepared_inputs: BatchEncoding,
    ) -> torch.Tensor:
        if batched_inputs is None:
            batched_inputs = {"input": []}

            if "attention_mask" in prepared_inputs:
                batched_inputs["attention_mask"] = []
        
        else:
            batched_inputs = batched_inputs[0]

        batched_inputs["input"].extend(prepared_inputs["input"])

        if "attention_mask" in prepared_inputs:
            batched_inputs["attention_mask"].extend(prepared_inputs["attention_mask"])

        return (batched_inputs,)

    def _execute_forward(self, prepared_inputs, *args, **kwargs) -> Any:

        # HookedTransformer uses attention_mask in forward but not in generate.
        if "attention_mask" in prepared_inputs:
            prepared_inputs.pop("attention_mask")

        return super()._execute_forward(prepared_inputs, *args, **kwargs)

    def _execute_generate(self, prepared_inputs, *args, **kwargs) -> Any:

        # HookedTransformer uses attention_mask in forward but not in generate.
        if "attention_mask" in prepared_inputs:
            prepared_inputs.pop("attention_mask")

        return super()._execute_generate(prepared_inputs, *args, **kwargs)



from typing_extensions import Self

from ... import NNsight
from ...util import from_import_path, to_import_path


class RemoteableMixin(NNsight):

    def _remoteable_model_key(self) -> str:

        raise NotImplementedError()

    @classmethod
    def _remoteable_from_model_key(cls, model_key: str) -> Self:
        raise NotImplementedError()

    def to_model_key(self):

        return f"{to_import_path(type(self))}:{self._remoteable_model_key()}"

    @classmethod
    def from_model_key(cls, model_key: str, **kwargs) -> Self:

        import_path, model_key = model_key.split(":", 1)

        type: RemoteableMixin = from_import_path(import_path)

        return type._remoteable_from_model_key(model_key, **kwargs)



from collections import OrderedDict, defaultdict
from typing import TYPE_CHECKING, Dict, List, Optional

from . import protocols

if TYPE_CHECKING:
    from ..intervention import InterventionProxy
    from .Graph import Graph
    from .Node import Node


class Bridge:
    """A Bridge object collects and tracks multiple Graphs in order to facilitate interaction between them.
    The order in which Graphs added matters as Graphs can only get values from previous Graphs/

    Attributes:
        id_to_graph (Dict[int, Graph]): Mapping of graph id to Graph.
        graph_stack (List[Graph]): Stack of visited Intervention Graphs.
        bridged_nodes (defaultdict[Node, defaultdict[int, Optional[InterventionProxy]]]): Mapping of bridged Nodes 
            to the BridgeProtocol nodes representing them on different graphs. 
        locks (int): Count of how many entities are depending on ties between graphs not to be released.
    """

    def __init__(self) -> None:

        # Mapping fro Graph if to Graph.
        self.id_to_graph: Dict[int, "Graph"] = OrderedDict()
        # Stack to keep track of most inner current graph
        self.graph_stack: List["Graph"] = list()
        self.bridged_nodes: defaultdict["Node", defaultdict[int, "InterventionProxy"]] = defaultdict(lambda: defaultdict(lambda: None))

        self.locks = 0

    @property
    def release(self) -> bool:

        return not self.locks

    def add(self, graph: "Graph") -> None:
        """Adds Graph to Bridge.

        Args:
            graph (Graph): Graph to add.
        """

        protocols.BridgeProtocol.set_bridge(graph, self)

        self.id_to_graph[graph.id] = graph

        self.graph_stack.append(graph)

    def peek_graph(self) -> "Graph":
        """Gets the current hierarchical Graph in the Bridge.

        Returns:
            Graph: Graph of current context.

        """
        return self.graph_stack[-1]

    def pop_graph(self) -> None:
        """Pops the last Graph in the graph stack."""

        self.graph_stack.pop()

    def get_graph(self, id: int) -> "Graph":
        """Returns graph from Bridge given the Graph's id.

        Args:
            id (int): Id of Graph to get.

        Returns:
            Graph: Graph.
        """

        return self.id_to_graph[id]
    
    def add_bridge_proxy(self, node: "Node", bridge_proxy: "Node") -> None:
        """ Adds a BridgeProtocol Proxy to the bridged nodes attribute.

        Args:
            - node (Node): Bridged Node.
            - bridge_proxy (Node): BridgeProtocol node proxy corresponding to the bridged node.
        """ 

        self.bridged_nodes[node][bridge_proxy.node.graph.id] = bridge_proxy

    def get_bridge_proxy(self, node: "Node", graph_id: int) -> Optional["InterventionProxy"]:
        """ Check if the argument Node is bridged within the specified graph and returns its corresponding BridgeProtocol node proxy.

        Args:
            - node (Node): Node.
            - graph_id (int): Graph id.

        Returns: 
            Optional[InterventionProxy]: BridgeProtocol node proxy if it exists.
        """

        return self.bridged_nodes[node][graph_id]



from typing import Iterator

import torch
from torch.nn.parameter import Parameter
from typing_extensions import Self

from .tracing.Proxy import Proxy


class Module(torch.nn.Module):

    def save(self) -> Self:

        [param.save() for param in self.parameters()]

        return self

    def parameters(self, recurse: bool = True) -> Iterator[Parameter]:
        return [
            param
            for param in self.__dict__.values()
            if isinstance(param, Proxy)
            and param.node.target is torch.nn.parameter.Parameter
        ]



import operator
from inspect import getmembers, isbuiltin, isfunction, ismethod, ismethoddescriptor, isclass

import einops
import torch
from torch.utils.data.dataloader import DataLoader

from ... import intervention, util
from ...tracing import protocols
from ...tracing.Proxy import Proxy


def get_function_name(fn, module_name=None):
    if isinstance(fn, str):
        return fn

    if module_name is not None:
        return f"{module_name}.{fn.__name__}"

    module_name = getattr(fn, "__module__", None)

    if module_name is None:
        return fn.__qualname__

    return f"{module_name}.{fn.__qualname__}"


FUNCTIONS_WHITELIST = {}

### Torch functions
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(torch._C._VariableFunctions, isbuiltin)
    }
)
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(torch.nn.functional, isfunction)
    }
)
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(torch._C._nn, isfunction)
    }
)
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value, module_name="Tensor"): value
        for key, value in getmembers(
            torch.Tensor, lambda x: ismethoddescriptor(x) or isfunction(x)
        )
    }
)
FUNCTIONS_WHITELIST.update({get_function_name(torch.nn.Parameter): torch.nn.Parameter})

FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(torch.optim, isclass)
        if issubclass(value, torch.optim.Optimizer)
    }
)
FUNCTIONS_WHITELIST.update({get_function_name(DataLoader): DataLoader})
### operator functions
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(operator, isbuiltin)
        if not key.startswith("_")
    }
)
### einops functions
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(value): value
        for key, value in getmembers(einops.einops, isfunction)
    }
)

### nnsight functions
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(bool): bool,
        get_function_name(bytes): bytes,
        get_function_name(int): int,
        get_function_name(float): float,
        get_function_name(str): str,
        get_function_name(complex): complex,
        get_function_name(bytearray): bytearray,
        get_function_name(tuple): tuple,
        get_function_name(list): list,
        get_function_name(set): set,
        get_function_name(dict): dict,
        get_function_name(print): print,
        get_function_name(setattr): setattr,
        get_function_name(util.fetch_attr): util.fetch_attr,
        get_function_name(Proxy.proxy_call): Proxy.proxy_call,
    }
)

### protocols
FUNCTIONS_WHITELIST.update(
    {
        get_function_name(protocol): protocol
        for key, protocol in getmembers(protocols)
        if isinstance(protocol, type) and issubclass(protocol, protocols.Protocol)
    }
)

FUNCTIONS_WHITELIST.update(
    {
        get_function_name(protocol): protocol
        for key, protocol in getmembers(intervention)
        if isinstance(protocol, type) and issubclass(protocol, protocols.Protocol)
    }
)



from __future__ import annotations

import inspect
import operator
import weakref
from typing import TYPE_CHECKING, Any, Callable, Union

from typing_extensions import Self

from .. import util

if TYPE_CHECKING:
    from .Node import Node


class Proxy:
    """Proxy objects are the actual objects that interact with operations in order to update the graph to create new Nodes.

    The operations that are traceable on base Proxy objects are many python built-in and magic methods, as well as implementing __torch_function__ to trace torch operations.

    Attributes:
        node (Node): This proxy's node.
    """

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d: dict):
        self.__dict__ = d

    @staticmethod
    def proxy_call(callable: Callable, *args, **kwargs) -> Self:
        return callable(*args, **kwargs)

    def __init__(self, node: "Node") -> None:

        self.__dict__["node"] = node

        self.node: "Node"
        self.node.proxy = weakref.proxy(self)

    @property
    def value(self) -> Any:
        """Property to return the value of this proxy's node.

        Returns:
            Any: The stored value of the proxy, populated during execution of the model.
        """

        return self.node.value

    def __str__(self) -> str:

        if not self.node.attached():

            return str(self.value)

        return f"{type(self).__name__} ({self.node.name}): {self.node.proxy_value if self.node.proxy_value is not inspect._empty else ''}"

    def __repr__(self) -> str:

        if not self.node.attached():

            return repr(self.value)

        return str(self)

    def __call__(self, *args, **kwargs) -> Self:
        """
        Calling a Proxy object just creates a Proxy.proxy_call operation.

        Returns:
            Proxy: New call proxy.
        """

        return self.node.create(
            target=Proxy.proxy_call,
            args=[self.node] + list(args),
            kwargs=kwargs,
        )

    def __getitem__(self, key: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.getitem,
            args=[self.node, key],
        )

    def __setitem__(self, key: Union[Proxy, Any], value: Union[Self, Any]) -> None:
        self.node.create(
            target=operator.setitem,
            args=[self.node, key, value],
        )

    def __getattr__(self, key: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=util.fetch_attr,
            args=[self.node, key],
        )

    def __setattr__(self, key: Union[Proxy, Any], value: Union[Self, Any]) -> None:

        if key == "__dict__":

            super().__setattr__(key, value)

            return

        return self.node.create(
            target=setattr,
            args=[self.node, key, value],
        )

    def __len__(self) -> Self:
        return self.node.create(
            target=len,
            args=[self.node],
        )

    def __abs__(self) -> Self:
        return self.node.create(
            target=operator.abs,
            args=[self.node],
        )

    def __invert__(self) -> Self:
        return self.node.create(
            target=operator.invert,
            args=[self.node],
        )

    def __neg__(self) -> Self:
        return self.node.create(
            target=operator.neg,
            args=[self.node],
        )

    def __index__(self) -> Self:
        return self.node.create(target=operator.index, args=[self.node])

    def __add__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.add,
            args=[self.node, other],
        )

    def __radd__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.add,
            args=[other, self.node],
        )

    def __sub__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.sub,
            args=[self.node, other],
        )

    def __rsub__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.sub,
            args=[other, self.node],
        )

    def __pow__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.pow,
            args=[self.node, other],
        )

    def __rpow__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.pow,
            args=[other, self.node],
        )

    def __mul__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.mul,
            args=[self.node, other],
        )

    def __rmul__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.mul,
            args=[other, self.node],
        )

    def __mod__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.mod,
            args=[self.node, other],
        )

    def __rmod__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.mod,
            args=[other, self.node],
        )

    def __matmul__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.matmul,
            args=[self.node, other],
        )

    def __rmatmul__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.matmul,
            args=[other, self.node],
        )

    def __truediv__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.truediv,
            args=[self.node, other],
        )

    def __rtruediv__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(
            target=operator.truediv,
            args=[other, self.node],
        )

    def __floordiv__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.add(
            target=operator.floordiv,
            args=[self.node, other],
        )

    def __rfloordiv__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.add(
            target=operator.floordiv,
            args=[other, self.node],
        )

    def __eq__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.eq, args=[self.node, other])

    def __ne__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.ne, args=[self.node, other])

    def __lt__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.lt, args=[self.node, other])

    def __gt__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.gt, args=[self.node, other])

    def __le__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.le, args=[self.node, other])

    def __ge__(self, other: Union[Proxy, Any]) -> Self:
        return self.node.create(target=operator.ge, args=[self.node, other])

    def __index__(self) -> Self:
        return self.node.create(target=operator.index, args=[self.node])

    def __bool__(self) -> bool:
        return self.node.proxy_value.__bool__()

    def __instancecheck__(self, __instance: Any) -> bool:
        return self.node.proxy_value.__instancecheck__(__instance)

    @classmethod
    def __torch_function__(cls, orig_method, types, args=None, kwargs=None) -> Self:
        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()

        proxy: Proxy = None

        def get_proxy(arg):
            nonlocal proxy

            proxy = arg

        util.apply(args, get_proxy, Proxy)

        return proxy.node.create(
            target=orig_method,
            args=args,
            kwargs=kwargs,
        )


from functools import wraps


def proxy_wrapper(fn) -> Callable:
    """Wraps problematic functions (torch functions sometimes).
    Checks if any of its args are proxies. If so we return a proxy of the function.
    Otherwise just run the function.

    Args:
        fn (function): Function to wrap.

    Returns:
        function: Wrapped function.
    """

    @wraps(fn)
    def patched(*args, **kwargs):

        node = None

        def get_node(proxy: Proxy):
            nonlocal node

            node = proxy.node

        util.apply((args, kwargs), get_node, Proxy)

        if node is not None:
            return node.create(target=fn, args=args, kwargs=kwargs)

        else:
            return fn(*args, **kwargs)

    return patched



from __future__ import annotations

from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Dict, List, Optional, Set, Any, Union

from ..tracing import protocols

if TYPE_CHECKING:
    from ..tracing.Node import Node
    from ..tracing.Graph import Graph
    from ..intervention import InterventionProxy

class ConditionalManager():
    """ A Graph attachement that manages the Conditional contexts defined within an Intervention Graph.
    
    Attributes:
        _conditional_dict (Dict[str, Node]): Mapping of ConditionalProtocol node name to Conditional context.
        _conditioned_nodes_dict (Dict[str, Set[Node]]): Mapping of ConditionalProtocol node name to all the Nodes conditiones by it.
        _conditional_stack (Dict): Stack of visited Conditional contexts' ConditonalProtocol nodes.
    """

    def __init__(self):
        self._conditional_nodes_dict: Dict[str, Node] = dict()
        self._conditioned_nodes_dict: Dict[str, Set[Node]] = dict()
        self._conditional_nodes_stack: List[Node] = list()

    def push(self, conditional_node: "Node") -> None:
        """ Adds the Conditional to the stack of Conditional contexts.
        
        Args:
            conditional_node (Node): ConditionalProtocol node.
        """

        self._conditional_nodes_dict[conditional_node.name] = conditional_node
        self._conditioned_nodes_dict[conditional_node.name] = set()
        self._conditional_nodes_stack.append(conditional_node)

    def get(self, key: str) -> Conditional:
        """ Returns a ConditionalProtocol node.

        Args:
            key (str): ConditionalProtocol node name.

        Returns:    
            Node: ConditionalProtocol node.
        """

        return self._conditional_nodes_dict[key]

    def pop(self) -> None:
        """ Pops the ConditionalProtocol node of the current Conditional context from the ConditionalManager stack. """

        self._conditional_nodes_stack.pop()

    def peek(self) -> Optional["Node"]:
        """ Gets the current Conditional context's ConditionalProtocol node.
        
        Returns:
            Optional[Node]: Lastest ConditonalProtocol node if the ConditionalManager stack is non-empty.
        """

        if len(self._conditional_nodes_stack) > 0:
            return self._conditional_nodes_stack[-1]
        
    def add_conditioned_node(self, node: "Node") -> None:
        """  Adding a Node to the set of conditioned nodes by the current Conditonal context.

        Args:
            - node (Node): A node conditioned by the latest Conditional context.
        """

        self._conditioned_nodes_dict[self.peek().name].add(node)

    def is_node_conditioned(self, node: "Node") -> bool:
        """ Returns True if the Node argument is conditioned by the current Conditional context.
        
        Args:
            - node (Node): Node.

        Returns:
            bool: Whether the Node is conditioned.
        """

        curr_conditioned_nodes_set = self._conditioned_nodes_dict[self.peek().name]

        return (node in curr_conditioned_nodes_set)
    

class Conditional(AbstractContextManager):
    """ A context defined by a boolean condition, upon which the execution of all nodes defined from within is contingent. 

    Attributes:
        _graph (Graph): Conditional Context graph.
        _condition (Union[InterventionProxy, Any]): Condition.
    """

    def __init__(self, graph: "Graph", condition: Union["InterventionProxy", Any]):
       self._graph = graph
       self._condition: Union["InterventionProxy", Any] = condition

    def __enter__(self) -> Conditional:

        conditional_node = protocols.ConditionalProtocol.add(self._graph, self._condition).node

        protocols.ConditionalProtocol.push_conditional(conditional_node)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:

        protocols.ConditionalProtocol.pop_conditional(self._graph)



from __future__ import annotations

from typing import Any, Callable, Dict, List, Tuple, Union

import accelerate
import causal_conv1d_cuda
import mamba_ssm
import selective_scan_cuda
import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from mamba_ssm import MambaLMHeadModel
from mamba_ssm.models.config_mamba import MambaConfig
from mamba_ssm.utils.hf import load_config_hf, load_state_dict_hf
from transformers import AutoTokenizer, BatchEncoding, PreTrainedModel

from nnsight.util import WrapperModule

from ..patching import Patch, Patcher
from .LanguageModel import LanguageModel

from torch._guards import detect_fake_mode

class Mamba(LanguageModel):

    def _load(self, repo_id: str, device='cpu', **kwargs) -> PreTrainedModel:

        config = MambaConfig(**load_config_hf(repo_id))

        if self.tokenizer is None:

            self.tokenizer = AutoTokenizer.from_pretrained(
                repo_id, config=config, padding_side="left"
            )
            self.tokenizer.pad_token = self.tokenizer.eos_token

        if self._model is None:
            
            model = MambaLMHeadModel(config, device="meta", dtype=None, **kwargs)
            
            setattr(model, 'generator', WrapperModule())

            return model

        model = MambaLMHeadModel(config, device=device, **kwargs)
        model.load_state_dict(load_state_dict_hf(repo_id, device=device, **kwargs))
        
        setattr(model, 'generator', WrapperModule())

        return model
    
    def _execute_forward(self, prepared_inputs: Any, *args, **kwargs):

        device = next(self._model.parameters()).device

        patcher = None

        with Patcher() as patcher:

            if detect_fake_mode(prepared_inputs):

                def blah(hs, *args, residual=None, **kwargs):
                    return hs, residual or torch.rand_like(hs)

                def blah1(hs, *args, **kwargs):
                    return hs

                def blah2(hs, *args, **kwargs):
                    return hs

                def blah3(conv1d_out, delta, A, B, C, D, z, delta_bias, delta_softplus):
                    return (
                        conv1d_out,
                        torch.zeros((*conv1d_out.shape, A.shape[1] * 2), device="meta"),
                        conv1d_out,
                    )

                
                patcher.add(Patch(mamba_ssm.modules.mamba_simple, blah, "rms_norm_fn"))
                patcher.add(
                    Patch(mamba_ssm.models.mixer_seq_simple, blah1, "rms_norm_fn")
                )
                patcher.add(Patch(causal_conv1d_cuda, blah2, "causal_conv1d_fwd"))
                patcher.add(Patch(selective_scan_cuda, blah3, "fwd"))


            return self._model(
                prepared_inputs["input_ids"].to(device),
                *args,
                **kwargs,
            )

    def _execute_generate(
        self, prepared_inputs: Any, *args, max_length=1, **kwargs
    ):

        device = next(self._model.parameters()).device

        patcher = None

        with Patcher() as patcher:

            if detect_fake_mode(prepared_inputs):

                def blah(hs, *args, residual=None, **kwargs):
                    return hs, residual or torch.rand_like(hs)

                def blah1(hs, *args, **kwargs):
                    return hs

                def blah2(hs, *args, **kwargs):
                    return hs

                def blah3(conv1d_out, delta, A, B, C, D, z, delta_bias, delta_softplus):
                    return (
                        conv1d_out,
                        torch.zeros((*conv1d_out.shape, A.shape[1] * 2), device="meta"),
                        conv1d_out,
                    )

                
                patcher.add(Patch(mamba_ssm.modules.mamba_simple, blah, "rms_norm_fn"))
                patcher.add(
                    Patch(mamba_ssm.models.mixer_seq_simple, blah1, "rms_norm_fn")
                )
                patcher.add(Patch(causal_conv1d_cuda, blah2, "causal_conv1d_fwd"))
                patcher.add(Patch(selective_scan_cuda, blah3, "fwd"))

            output = self._model.generate(
                prepared_inputs["input_ids"].to(device),
                *args,
                max_length=max_length,
                **kwargs,
            )

        self._model.generator(output)

        return output

class SSM(torch.nn.Module):
    class DiscA(torch.nn.Module):
        def forward(self, delta, A):
            return torch.exp(torch.einsum("bdl,dn->bdln", delta, A))

    class DiscB(torch.nn.Module):
        def forward(self, delta, B):
            return torch.einsum("bdl,bnl->bdln", delta, B)

    class Hx(torch.nn.Module):
        class Bx(torch.nn.Module):
            def forward(self, deltaB: torch.Tensor, x: torch.Tensor):
                return torch.einsum("bdn,bd->bdn", deltaB, x)

        class Ah(torch.nn.Module):
            def forward(self, deltaA: torch.Tensor, h: torch.Tensor):
                return deltaA * h

        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)

            self.bx = SSM.Hx.Bx()
            self.ah = SSM.Hx.Ah()

        def forward(
            self,
            deltaA: torch.Tensor,
            deltaB: torch.Tensor,
            x: torch.Tensor,
            h: torch.Tensor,
        ):
            return self.ah(deltaA, h) + self.bx(deltaB, x)

    class Yh(torch.nn.Module):
        def forward(self, h, C):
            y = torch.einsum("bdn,bn->bd", h, C)

            if y.is_complex():
                y = y.real * 2

            return y

    def __init__(self):
        super().__init__()

        self.discA = SSM.DiscA()
        self.discB = SSM.DiscB()
        self.hx = SSM.Hx()
        self.yh = SSM.Yh()

    def forward(self, x, delta, A, B, C, D=None, z=None, return_last_state=False):
        dtype_in = x.dtype

        x = x.float()
        delta = delta.float()

        batch, dim, dstate = x.shape[0], A.shape[0], A.shape[1]

        if A.is_complex():
            B = torch.view_as_complex(
                rearrange(B.float(), "... (L two) -> ... L two", two=2)
            )
            C = torch.view_as_complex(
                rearrange(C.float(), "... (L two) -> ... L two", two=2)
            )
        else:
            B = B.float()
            C = C.float()

        deltaA = self.discA(delta, A)

        deltaB = self.discB(delta, B)

        last_state = None

        h = A.new_zeros((batch, dim, dstate))

        ys = []

        # Main recurrence loop
        for token_idx in range(x.shape[2]):
            h = self.hx(
                deltaA[:, :, token_idx], deltaB[:, :, token_idx], x[:, :, token_idx], h
            )

            y = self.yh(h, C[:, :, token_idx])

            if token_idx == x.shape[2] - 1:
                last_state = h

            ys.append(y)

        y = torch.stack(ys, dim=2)  # (batch dim L)

        out = y if D is None else y + x * rearrange(D, "d -> d 1")

        if z is not None:
            out = out * F.silu(z)

        out = out.to(dtype=dtype_in)

        return out if not return_last_state else (out, last_state)


class MambaModuleInterp(mamba_ssm.modules.mamba_simple.Mamba):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.dt = WrapperModule()
        self.B = WrapperModule()
        self.C = WrapperModule()

        self.ssm = SSM()

        self.delta_softplus = torch.nn.Softplus()

    def forward(self, hidden_states, inference_params=None):
        """
        hidden_states: (B, L, D)
        Returns: same shape as hidden_states
        """
        batch, seqlen, dim = hidden_states.shape

        conv_state, ssm_state = None, None
        if inference_params is not None:
            conv_state, ssm_state = self._get_states_from_cache(inference_params, batch)
            if inference_params.seqlen_offset > 0:
                # The states are updated inplace
                out, _, _ = self.step(hidden_states, conv_state, ssm_state)
                return out

        # We do matmul and transpose BLH -> HBL at the same time
        xz = rearrange(
            self.in_proj(hidden_states),
            "b l d -> b d l",
            l=seqlen,
        )

        A = -torch.exp(self.A_log.float())  # (d_inner, d_state)

        # In the backward pass we write dx and dz next to each other to avoid torch.cat
        x, z = xz.chunk(2, dim=1)

        # Compute short convolution
        if conv_state is not None:
            # If we just take x[:, :, -self.d_conv :], it will error if seqlen < self.d_conv
            # Instead F.pad will pad with zeros if seqlen < self.d_conv, and truncate otherwise.
            conv_state.copy_(
                F.pad(x, (self.d_conv - x.shape[-1], 0))
            )  # Update state (B D W)

        x = self.act(self.conv1d(x)[..., :seqlen])

        x_dbl = self.x_proj(rearrange(x, "b d l -> (b l) d"))  # (bl d)
        dt, B, C = torch.split(
            x_dbl, [self.dt_rank, self.d_state, self.d_state], dim=-1
        )
        dt = self.dt_proj(dt).t()

        dt = rearrange(dt, "d (b l) -> b d l", l=seqlen)
        B = rearrange(B, "(b l) dstate -> b dstate l", l=seqlen).contiguous()
        C = rearrange(C, "(b l) dstate -> b dstate l", l=seqlen).contiguous()

        dt = self.dt(dt)
        B = self.B(B)
        C = self.C(C)

        dt = self.delta_softplus(dt)

        y = self.ssm(
            x,
            dt,
            A,
            B,
            C,
            self.D.float(),
            z=z,
            return_last_state=ssm_state is not None,
        )
        if ssm_state is not None:
            y, last_state = y
            ssm_state.copy_(last_state)

        y = rearrange(y, "b d l -> b l d")

        out = self.out_proj(y)

        return out


class MambaInterp(Mamba):
    def __init__(self, *args, **kwargs):
        patcher = Patcher()

        patcher.add(
            Patch(
                mamba_ssm.models.mixer_seq_simple,
                MambaModuleInterp,
                "Mamba",
            )
        )

        patcher.__enter__()

        super().__init__(*args, **kwargs)



from __future__ import annotations

import inspect
import weakref
from contextlib import AbstractContextManager
from functools import wraps
from typing import Any, Callable, Union

import torch
from torch.overrides import TorchFunctionMode
from typing_extensions import Self

from ..intervention import InterventionProxy
from ..patching import Patch, Patcher
from ..tracing import protocols
from ..tracing.Bridge import Bridge
from ..tracing.Graph import Graph
from .backends import Backend, BridgeMixin
from .Conditional import Conditional


class GraphBasedContext(AbstractContextManager, BridgeMixin):

    def __init__(
        self,
        backend: Backend,
        graph: Graph = None,
        bridge: Bridge = None,
        **kwargs,
    ) -> None:

        self.backend = backend

        self.graph: Graph = Graph(**kwargs) if graph is None else graph

        if bridge is not None:

            bridge.add(self.graph)

    def apply(
        self,
        target: Callable,
        *args,
        validate: bool = None,
        **kwargs,
    ) -> InterventionProxy:
        """Helper method to directly add a function to the intervention graph.

        Args:
            target (Callable): Function to apply
            validate (bool): If to try and run this operation in FakeMode to test it out and scan it.

        Returns:
            InterventionProxy: Proxy of applying that function.
        """

        proxy_value = inspect._empty

        if validate is False:

            proxy_value = None

        return self.graph.create(
            target=target,
            proxy_value=proxy_value,
            args=args,
            kwargs=kwargs,
        )

    def cond(self, condition: Union[InterventionProxy, Any]) -> Conditional:
        """Entrypoint to the Conditional context.
            Takes in a condition argument which acts as the dependency of the Conditional node in the Intervention graph.
            The condition is evaluated as a boolean, and if True, executes all the interventions defined within the body
            of the conditional context.

        Args:
            condition (Union[InterventionProxy, Any]): Dependency of the Conditional Node.

        Returns:
            Conditional: Conditional context object.

        Example:

            Setup:
                .. code-block:: python
                    import torch
                    from collections import OrderedDict

                    input_size = 5
                    hidden_dims = 10
                    output_size = 2

                    model = nn.Sequential(OrderedDict([
                        ('layer1', torch.nn.Linear(input_size, hidden_dims)),
                        ('layer2', torch.nn.Linear(hidden_dims, output_size)),
                    ]))

                    input = torch.rand((1, input_size))

            Ex 1: The .save() on the model output will only be executed if the condition passed to tracer.cond() is evaluated to True.

            .. code-block:: python
                x: int = 5
                with model.trace(input) as trace:
                    with tracer.cond(x > 0):
                        out = model.output.save()

            Ex 2: The condition is on an InterventionProxy which creates in return an InterventionProxy

            .. code-block:: python
                with model.trace(input) as trace:
                    with tracer.cond(model.layer1.output[:, 0] > 0):
                        out = model.output.save()
        """

        return Conditional(self.graph, condition)

    def exit(self) -> InterventionProxy:
        """Exits the execution of a sequential intervention graph.

        Returns:
            InterventionProxy: Proxy of the EarlyStopProtocol node.
        """

        if self.graph.sequential:
            return protocols.EarlyStopProtocol.add(self.graph)
        else:
            raise Exception(
                "Early exit is only supported for sequential graph-based contexts."
            )

    def log(self, *data: Any) -> None:
        """Adds a node via .apply to print the value of a Node.

        Args:
            data (Any): Data to print.
        """
        self.apply(print, *data)

    def bool(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable bool."""

        return self.apply(bool, *args, **kwargs)

    def bytes(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable bytes."""

        return self.apply(bytes, *args, **kwargs)

    def int(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable int."""

        return self.apply(int, *args, **kwargs)

    def float(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable float."""

        return self.apply(float, *args, **kwargs)

    def str(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable string."""

        return self.apply(str, *args, **kwargs)

    def complex(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable complex number."""

        return self.apply(complex, *args, **kwargs)

    def bytearray(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable bytearray."""

        return self.apply(bytearray, *args, **kwargs)

    def tuple(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable tuple."""

        return self.apply(tuple, *args, **kwargs)

    def list(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable list."""

        return self.apply(list, *args, **kwargs)

    def set(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable set."""

        return self.apply(set, *args, **kwargs)

    def dict(self, *args, **kwargs) -> InterventionProxy:
        """NNsight helper method to create a traceable dictionary."""

        return self.apply(dict, *args, **kwargs)

    def vis(self, **kwargs) -> None:
        """
        Helper method to save a visualization of the current state of the intervention graph.
        """

        self.graph.vis(**kwargs)

    def __enter__(self) -> Self:

        GlobalTracingContext.try_register(self)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:

        GlobalTracingContext.try_deregister(self)

        if isinstance(exc_val, BaseException):
            self.graph.alive = False
            self.graph = None
            raise exc_val

        self.backend(self)

    ### BACKENDS ########

    def local_backend_execute(self) -> None:

        try:
            self.graph.reset()
            self.graph.execute()
        except protocols.EarlyStopProtocol.EarlyStopException as e:
            raise e
        finally:
            graph = self.graph
            graph.alive = False

            if not isinstance(graph, weakref.ProxyType):
                self.graph = weakref.proxy(graph)

    def bridge_backend_handle(self, bridge: Bridge) -> None:

        bridge.pop_graph()

        protocols.LocalBackendExecuteProtocol.add(self, bridge.peek_graph())

        self.graph = weakref.proxy(self.graph)


from inspect import getmembers, isclass

from torch.utils import data


def global_patch(root, name: str) -> Patch:

    fn = getattr(root, name)

    @wraps(fn)
    def inner(*args, **kwargs):

        return GlobalTracingContext.GLOBAL_TRACING_CONTEXT.apply(fn, *args, **kwargs)

    return Patch(root, inner, name)


def global_patch_class(cls: type) -> Patch:

    fn = cls.__new__

    @wraps(fn)
    def inner(cls, *args, **kwargs):

        return GlobalTracingContext.GLOBAL_TRACING_CONTEXT.apply(cls, *args, **kwargs)

    return Patch(cls, inner, "__new__")


class GlobalTracingContext(GraphBasedContext):
    """The Global Tracing Context handles adding tracing operations globally without reference to a given `GraphBasedContext`.
    There should only be one of these and that is `GlobalTracingContext.GLOBAL_TRACING_CONTEXT`.
    `GlobalTracingContext.TORCH_HANDLER` handles adding torch functions without reference to a given `GraphBasedContext`.

    """

    GLOBAL_TRACING_CONTEXT: GlobalTracingContext
    TORCH_HANDLER: GlobalTracingContext.GlobalTracingTorchHandler
    PATCHER: Patcher = Patcher(
        [
            global_patch_class(torch.nn.Parameter),
            global_patch_class(data.DataLoader),
            global_patch(torch, "arange"),
            global_patch(torch, "empty"),
            global_patch(torch, "eye"),
            global_patch(torch, "full"),
            global_patch(torch, "linspace"),
            global_patch(torch, "logspace"),
            global_patch(torch, "ones"),
            global_patch(torch, "rand"),
            global_patch(torch, "randint"),
            global_patch(torch, "randn"),
            global_patch(torch, "randperm"),
            global_patch(torch, "zeros"),
        ]
        + [
            global_patch_class(value)
            for key, value in getmembers(torch.optim, isclass)
            if issubclass(value, torch.optim.Optimizer)
        ]
    )

    class GlobalTracingTorchHandler(TorchFunctionMode):

        def __torch_function__(self, func, types, args, kwargs=None):

            if kwargs is None:

                kwargs = {}

            if "_VariableFunctionsClass" in func.__qualname__:
                return GlobalTracingContext.GLOBAL_TRACING_CONTEXT.apply(
                    func, *args, **kwargs
                )

            return func(*args, **kwargs)

    class GlobalTracingExit(AbstractContextManager):

        def __enter__(self) -> Any:

            GlobalTracingContext.TORCH_HANDLER.__exit__(None, None, None)
            GlobalTracingContext.PATCHER.__exit__(None, None, None)

            return self

        def __exit__(self, exc_type, exc_val, traceback):

            GlobalTracingContext.TORCH_HANDLER.__enter__()
            GlobalTracingContext.PATCHER.__enter__()

            if isinstance(exc_val, BaseException):

                raise exc_val

    def __init__(self) -> None:
        """We create an empty `GraphBasedContext` by default."""

        self.graph: Graph = None

    @staticmethod
    def exit_global_tracing_context():

        return GlobalTracingContext.GlobalTracingExit()

    @staticmethod
    def try_register(graph_based_context: GraphBasedContext) -> bool:
        """Attempts to register a `Graph` globally.]
        Will not if one is already registered.

        Args:
            graph_based_context (GraphBasedContext): `GraphBasedContext` to register.

        Returns:
            bool: True if registering ws successful, False otherwise.
        """

        if GlobalTracingContext.GLOBAL_TRACING_CONTEXT:

            return False

        GlobalTracingContext.register(graph_based_context)

        return True

    @staticmethod
    def try_deregister(graph_based_context: GraphBasedContext) -> bool:
        """Attempts to deregister a `Graph` globally.
        Will not if `graph_based_context` does not have the same `Graph` as the currently registered one.

        Args:
            graph_based_context (GraphBasedContext): `GraphBasedContext` to deregister.

        Returns:
            bool: True if deregistering ws successful, False otherwise.
        """
        if (
            not GlobalTracingContext.GLOBAL_TRACING_CONTEXT
            or graph_based_context.graph
            is not GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph
        ):

            return False
        GlobalTracingContext.deregister()

        return True

    @staticmethod
    def register(graph_based_context: GraphBasedContext) -> None:
        """Register `GraphBasedContext` globally.

        Args:
            graph_based_context (GraphBasedContext): GraphBasedContext to register.
        """

        assert GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph is None

        GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph = graph_based_context.graph

        GlobalTracingContext.TORCH_HANDLER.__enter__()
        GlobalTracingContext.PATCHER.__enter__()

    @staticmethod
    def deregister() -> None:
        """Deregister `GraphBasedContext` globally.

        Args:
            graph_based_context (GraphBasedContext): GraphBasedContext to deregister.
        """

        assert GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph is not None

        GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph = None

        GlobalTracingContext.TORCH_HANDLER.__exit__(None, None, None)
        GlobalTracingContext.PATCHER.__exit__(None, None, None)

    def __bool__(self) -> bool:
        """True if there is a `GraphBasedContext` registered globally. False otherwise."""

        return GlobalTracingContext.GLOBAL_TRACING_CONTEXT.graph is not None

    def __getattribute__(self, name: str) -> Any:
        """Prevent attribute access if no `GraphBasedContext` registered."""

        static_methods = [
            name
            for name, value in inspect.getmembers(
                GraphBasedContext, predicate=inspect.ismethod
            )
        ]

        if name in static_methods:

            if not GlobalTracingContext.GLOBAL_TRACING_CONTEXT:

                raise Exception(
                    "Global ops cannot be used outside of a tracing context."
                )

        return object.__getattribute__(self, name)


GlobalTracingContext.GLOBAL_TRACING_CONTEXT = GlobalTracingContext()
GlobalTracingContext.TORCH_HANDLER = GlobalTracingContext.GlobalTracingTorchHandler()



import torch
from abc import ABC, abstractmethod

from transformations import RotateLayer
from interventions_utils import sigmoid_boundary


class Intervention(torch.nn.Module, ABC):

    """Intervention the original representations."""
    def __init__(self):
        super().__init__()
        self.trainble = False
        
    @abstractmethod
    def set_interchange_dim(self, interchange_dim):
        pass

    @abstractmethod
    def forward(self, base, source):
        pass

class TrainbleIntervention(Intervention):

    """Intervention the original representations."""
    def __init__(self):
        super().__init__()
        self.trainble = True


class BoundlessRotatedSpaceIntervention(TrainbleIntervention):
    
    """Intervention in the rotated space with boundary mask."""
    def __init__(self, embed_dim, **kwargs):
        super().__init__()
        rotate_layer = RotateLayer(embed_dim)

        # Orthogonal parametrizations constrains the weights of the rotate layer to be 
        # orthogonal during training
        self.rotate_layer = torch.nn.utils.parametrizations.orthogonal(
            rotate_layer)

        self.intervention_boundaries = torch.nn.Parameter(
            torch.tensor([0.5]), requires_grad=True)
        self.temperature = torch.nn.Parameter(torch.tensor(50.0)) 
        self.embed_dim = embed_dim
        self.intervention_population = torch.nn.Parameter(
            torch.arange(0, self.embed_dim), requires_grad=False)
        
    def get_boundary_parameters(self):
        return self.intervention_boundaries

    # temporarily in here
    def zero_grad(self):
        """Zero out the gradients of all parameters in this module."""
        for param in self.parameters():
            if param.grad is not None:
                param.grad.zero_()

    def get_temperature(self):
        return self.temperature

    def set_temperature(self, temp: torch.Tensor):
        self.temperature.data = temp
        
    def set_interchange_dim(self, interchange_dim):
        """interchange dim is learned and can not be set"""
        assert False

    def forward(self, base, source):
        batch_size = base.shape[0]
        rotated_base = self.rotate_layer(base)
        rotated_source = self.rotate_layer(source)
        # get boundary
        intervention_boundaries = torch.clamp(
            self.intervention_boundaries, 1e-3, 1)
        boundary_mask = sigmoid_boundary(
            self.intervention_population.repeat(batch_size, 1), 
            0.,
            intervention_boundaries[0] * int(self.embed_dim),
            self.temperature
        )

        # print(boundary_mask.get_device())
        boundary_mask = torch.ones(
            batch_size).unsqueeze(dim=-1).to(boundary_mask.device)*boundary_mask
        # boundary_mask = boundary_mask.to(rotated_base.dtype)
        # interchange
        rotated_output = (1. - boundary_mask)*rotated_base + boundary_mask*rotated_source
        # inverse output
        output = torch.matmul(rotated_output, self.rotate_layer.weight.T)
        # return output.to(base.dtype)
        return output
    
    def __str__(self):
        return f"BoundlessRotatedSpaceIntervention(embed_dim={self.embed_dim})"


from __future__ import annotations

import weakref
from types import BuiltinFunctionType
from types import FunctionType as FuncType
from types import MethodDescriptorType
from typing import Dict, List, Literal, Optional, Union

import torch
from pydantic import BaseModel, ConfigDict, Field, Strict, field_validator
from pydantic.functional_validators import AfterValidator
from typing_extensions import Annotated

from ...contexts.session.Iterator import Iterator
from ...contexts.session.Session import Session
from ...contexts.Tracer import Tracer
from ...models.NNsightModel import NNsight
from ...tracing import protocols
from ...tracing.Bridge import Bridge
from ...tracing.Graph import Graph
from ...tracing.Node import Node
from . import FUNCTIONS_WHITELIST, get_function_name


class DeserializeHandler:

    def __init__(
        self,
        graph: Graph = None,
        nodes: Dict[str, Union[NodeModel, NodeType]] = None,
        model: NNsight = None,
        bridge: Bridge = None,
    ) -> None:

        self.graph = graph
        self.nodes = nodes
        self.model = model
        self.bridge = bridge


FUNCTION = Union[BuiltinFunctionType, FuncType, MethodDescriptorType, type]
PRIMITIVE = Union[int, float, str, bool, None]


class BaseNNsightModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    type_name: Literal["TYPE_NAME"]

    def deserialize(self, handler: DeserializeHandler):
        raise NotImplementedError()


### Custom Pydantic types for all supported base types
class NodeModel(BaseNNsightModel):

    type_name: Literal["NODE"] = "NODE"

    class Reference(BaseNNsightModel):
        type_name: Literal["NODE_REFERENCE"] = "NODE_REFERENCE"

        name: str

        def deserialize(self, handler: DeserializeHandler) -> Node:
            return handler.nodes[self.name].deserialize(handler)

    name: str
    target: Union[FunctionModel, FunctionType]
    args: List[ValueTypes]
    kwargs: Dict[str, ValueTypes]
    condition: Union[
        NodeReferenceType, NodeModel.Reference, PrimitiveModel, PrimitiveType
    ]

    def deserialize(self, handler: DeserializeHandler) -> Node:

        if self.name in handler.graph.nodes:
            return handler.graph.nodes[self.name]

        node = handler.graph.create(
            proxy_value=None,
            target=self.target.deserialize(handler),
            args=[value.deserialize(handler) for value in self.args],
            kwargs={
                key: value.deserialize(handler) for key, value in self.kwargs.items()
            },
            name=self.name,
        ).node

        node.cond_dependency = self.condition.deserialize(handler)
        if isinstance(node.cond_dependency, Node):
            node.cond_dependency.listeners.append(weakref.proxy(node))

        if isinstance(node.target, type) and issubclass(
            node.target, protocols.Protocol
        ):

            node.target.compile(node)

        return node


class PrimitiveModel(BaseNNsightModel):

    type_name: Literal["PRIMITIVE"] = "PRIMITIVE"

    value: PRIMITIVE

    def deserialize(self, handler: DeserializeHandler) -> PRIMITIVE:
        return self.value


class TensorModel(BaseNNsightModel):

    type_name: Literal["TENSOR"] = "TENSOR"

    values: List
    dtype: str

    def deserialize(self, handler: DeserializeHandler) -> torch.Tensor:
        dtype = getattr(torch, self.dtype)
        return torch.tensor(self.values, dtype=dtype)


class SliceModel(BaseNNsightModel):

    type_name: Literal["SLICE"] = "SLICE"

    start: ValueTypes
    stop: ValueTypes
    step: ValueTypes

    def deserialize(self, handler: DeserializeHandler) -> slice:

        return slice(
            self.start.deserialize(handler),
            self.stop.deserialize(handler),
            self.step.deserialize(handler),
        )


class EllipsisModel(BaseNNsightModel):

    type_name: Literal["ELLIPSIS"] = "ELLIPSIS"

    def deserialize(
        self, handler: DeserializeHandler
    ) -> type(
        ...
    ):  # It will be better to use EllipsisType, but it requires python>=3.10
        return ...


class ListModel(BaseNNsightModel):

    type_name: Literal["LIST"] = "LIST"

    values: List[ValueTypes]

    def deserialize(self, handler: DeserializeHandler) -> list:
        return [value.deserialize(handler) for value in self.values]


class TupleModel(BaseNNsightModel):

    type_name: Literal["TUPLE"] = "TUPLE"

    values: List[ValueTypes]

    def deserialize(self, handler: DeserializeHandler) -> tuple:
        return tuple([value.deserialize(handler) for value in self.values])


class DictModel(BaseNNsightModel):

    type_name: Literal["DICT"] = "DICT"

    values: Dict[str, ValueTypes]

    def deserialize(self, handler: DeserializeHandler) -> dict:
        return {key: value.deserialize(handler) for key, value in self.values.items()}


class FunctionWhitelistError(Exception):
    pass


class FunctionModel(BaseNNsightModel):

    type_name: Literal["FUNCTION"] = "FUNCTION"

    function_name: str

    @field_validator("function_name")
    @classmethod
    def check_function_whitelist(cls, qualname: str) -> str:
        if qualname not in FUNCTIONS_WHITELIST:
            raise FunctionWhitelistError(
                f"Function with name `{qualname}` not in function whitelist."
            )

        return qualname

    def deserialize(self, handler: DeserializeHandler) -> FUNCTION:
        return FUNCTIONS_WHITELIST[self.function_name]


class GraphModel(BaseNNsightModel):

    type_name: Literal["GRAPH"] = "GRAPH"

    id: int
    sequential: bool
    nodes: Dict[str, Union["NodeModel", "NodeType"]]

    def deserialize(self, handler: DeserializeHandler) -> Graph:

        graph = Graph(validate=False, sequential=self.sequential, graph_id=self.id)

        handler.graph = graph
        handler.nodes = self.nodes

        # To preserve order
        nodes = {}

        for node_name, node in self.nodes.items():

            node.deserialize(handler)

            # To preserve order
            nodes[node_name] = graph.nodes[node_name]

        # To preserve order
        graph.nodes = nodes

        return graph


class TracerModel(BaseNNsightModel):

    type_name: Literal["TRACER"] = "TRACER"

    kwargs: Dict[str, ValueTypes]
    invoker_inputs: List[ValueTypes]
    graph: Union[GraphModel, GraphType]

    def deserialize(self, handler: DeserializeHandler) -> Tracer:

        _graph = handler.graph
        _nodes = handler.nodes

        graph = self.graph.deserialize(handler)

        handler.graph = graph

        kwargs = {key: value.deserialize(handler) for key, value in self.kwargs.items()}

        invoker_inputs = [
            invoker_input.deserialize(handler) for invoker_input in self.invoker_inputs
        ]

        tracer = Tracer(
            None, handler.model, bridge=handler.bridge, graph=graph, **kwargs
        )
        tracer._invoker_inputs = invoker_inputs

        handler.graph = _graph
        handler.nodes = _nodes

        return tracer


class IteratorModel(BaseNNsightModel):

    type_name: Literal["ITERATOR"] = "ITERATOR"

    data: ValueTypes

    graph: Union[GraphModel, GraphType]

    def deserialize(self, handler: DeserializeHandler) -> Iterator:

        _graph = handler.graph
        _nodes = handler.nodes

        graph = self.graph.deserialize(handler)

        handler.graph = graph

        data = self.data.deserialize(handler)

        iterator = Iterator(data, None, bridge=handler.bridge, graph=graph)

        handler.graph = _graph
        handler.nodes = _nodes

        return iterator


class SessionModel(BaseNNsightModel):

    type_name: Literal["SESSION"] = "SESSION"

    graph: Union[GraphModel, GraphType]

    def deserialize(self, handler: DeserializeHandler) -> Session:

        bridge = Bridge()

        handler.bridge = bridge

        graph = self.graph.deserialize(handler)

        bridge.add(graph)

        session = Session(None, handler.model, bridge=bridge, graph=graph)

        return session


### Define Annotated types to convert objects to their custom Pydantic counterpart

GraphType = Annotated[
    Graph,
    AfterValidator(
        lambda value: GraphModel(
            id=value.id, sequential=value.sequential, nodes=value.nodes
        )
    ),
]

PrimitiveType = Annotated[
    PRIMITIVE, AfterValidator(lambda value: PrimitiveModel(value=value))
]

TensorType = Annotated[
    torch.Tensor,
    AfterValidator(
        lambda value: TensorModel(
            values=value.tolist(), dtype=str(value.dtype).split(".")[-1]
        )
    ),
]

SliceType = Annotated[
    slice,
    AfterValidator(
        lambda value: SliceModel(start=value.start, stop=value.stop, step=value.step)
    ),
]

EllipsisType = Annotated[
    type(...),  # It will be better to use EllipsisType, but it requires python>=3.10
    AfterValidator(lambda value: EllipsisModel()),
]


ListType = Annotated[list, AfterValidator(lambda value: ListModel(values=value))]

TupleType = Annotated[
    tuple, Strict(), AfterValidator(lambda value: TupleModel(values=list(value)))
]

DictType = Annotated[dict, AfterValidator(lambda value: DictModel(values=value))]

FunctionType = Annotated[
    FUNCTION,
    AfterValidator(lambda value: FunctionModel(function_name=get_function_name(value))),
]

NodeReferenceType = Annotated[
    Node, AfterValidator(lambda value: NodeModel.Reference(name=value.name))
]

NodeType = Annotated[
    Node,
    AfterValidator(
        lambda value: NodeModel(
            name=value.name,
            target=value.target,
            args=value.args,
            kwargs=value.kwargs,
            condition=value.cond_dependency,
        )
    ),
]

TracerType = Annotated[
    Tracer,
    AfterValidator(
        lambda value: TracerModel(
            kwargs=value._kwargs,
            invoker_inputs=value._invoker_inputs,
            graph=value.graph,
        )
    ),
]

IteratorType = Annotated[
    Iterator,
    AfterValidator(lambda value: IteratorModel(graph=value.graph, data=value.data)),
]

SessionType = Annotated[
    Session,
    AfterValidator(lambda value: SessionModel(graph=value.graph)),
]

### Register all custom Pydantic objects to convert objects to
TOTYPES = Union[
    TracerModel,
    IteratorModel,
    SessionModel,
    NodeModel.Reference,
    SliceModel,
    TensorModel,
    PrimitiveModel,
    TupleModel,
    ListModel,
    DictModel,
    EllipsisModel,
]
### Register all Annotated types objects to convert objects from
FROMTYPES = Union[
    TracerType,
    IteratorType,
    SessionType,
    NodeReferenceType,
    SliceType,
    TensorType,
    PrimitiveType,
    TupleType,
    ListType,
    DictType,
    EllipsisType,
]

### Final registration
ValueTypes = Union[
    Annotated[
        TOTYPES,
        Field(discriminator="type_name"),
    ],
    FROMTYPES,
]



from typing import Any

from . import Backend


class LocalMixin:
    """To be inherited by objects that want to be able to be executed by the LocalBackend."""

    def local_backend_execute(self) -> Any:
        """Should execute this object locally and return a result that can be handled by RemoteMixin objects.

        Returns:
            Any: Result containing data to return from a remote execution.
        """

        raise NotImplementedError()


class LocalBackend(Backend):
    """Backend to execute a context object on your local machine.

    Context object must inherit from LocalMixin and implement its methods.
    """

    def __call__(self, obj: LocalMixin):

        obj.local_backend_execute()



import torch

class RotateLayer(torch.nn.Module):
    """A linear transformation with orthogonal initialization."""

    def __init__(self, n, init_orth=True):
        super().__init__()
        weight = torch.empty(n, n)
        # we don't need init if the saved checkpoint has a nice
        # starting point already.
        # you can also study this if you want, but it is our focus.
        if init_orth:
            torch.nn.init.orthogonal_(weight)
        self.weight = torch.nn.Parameter(weight, requires_grad=True)

    def forward(self, x):
        return torch.matmul(x.to(self.weight.dtype), self.weight)




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
#                                                                                                                       #
#      ::::    ::: ::::    :::  :::::::: ::::::::::: ::::::::  :::    ::: :::::::::::       :::::::       ::::::::      #
#      :+:+:   :+: :+:+:   :+: :+:    :+:    :+:    :+:    :+: :+:    :+:     :+:          :+:   :+:     :+:    :+:     #
#      :+:+:+  +:+ :+:+:+  +:+ +:+           +:+    +:+        +:+    +:+     +:+          +:+  :+:+            +:+     #
#      +#+ +:+ +#+ +#+ +:+ +#+ +#++:++#++    +#+    :#:        +#++:++#++     +#+          +#+ + +:+         +#++:      #
#      +#+  +#+#+# +#+  +#+#+#        +#+    +#+    +#+   +#+# +#+    +#+     +#+          +#+#  +#+            +#+     #
#      #+#   #+#+# #+#   #+#+# #+#    #+#    #+#    #+#    #+# #+#    #+#     #+#          #+#   #+# #+# #+#    #+#     #
#      ###    #### ###    ####  ######## ########### ########  ###    ###     ###           #######  ###  ########      #
#                                                                                                                       #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
import os
from functools import wraps
from typing import Dict, Union

import torch
import yaml

from .patching import *
from .schema.Config import ConfigModel

PATH = os.path.dirname(os.path.abspath(__file__))
with open(os.path.join(PATH, "config.yaml"), "r") as file:
    CONFIG = ConfigModel(**yaml.safe_load(file))

from .logger import logger, remote_logger
from .models.NNsightModel import NNsight
from .models.LanguageModel import LanguageModel
from .patching import Patch, Patcher
from .tracing.Proxy import proxy_wrapper

logger.disabled = not CONFIG.APP.LOGGING
remote_logger.disabled = not CONFIG.APP.REMOTE_LOGGING

# Below do default patching:
DEFAULT_PATCHER = Patcher()

import math
from inspect import getmembers, isbuiltin, isfunction

import einops

for key, value in getmembers(einops.einops, isfunction):
    DEFAULT_PATCHER.add(Patch(einops.einops, proxy_wrapper(value), key))
for key, value in getmembers(math, isbuiltin):
    DEFAULT_PATCHER.add(Patch(math, proxy_wrapper(value), key))

# Tensor creation operations
from torch._subclasses.fake_tensor import FakeTensor


def _bool(self):
    return True


DEFAULT_PATCHER.add(Patch(FakeTensor, _bool, "__bool__"))


def fake_tensor_new_wrapper(fn):

    @wraps(fn)
    def inner(cls, fake_mode, elem, device, constant=None):

        if isinstance(elem, FakeTensor):

            return elem

        else:

            return fn(cls, fake_mode, elem, device, constant=constant)

    return inner


DEFAULT_PATCHER.add(
    Patch(FakeTensor, fake_tensor_new_wrapper(FakeTensor.__new__), "__new__")
)


def onehot_wrapper(fn):
    @wraps(fn)
    def onehot(input: torch.Tensor, num_classes=-1):
        if input.device.type == "meta":
            return torch.zeros((*input.shape, num_classes), device="meta")

        else:
            return fn(input, num_classes=num_classes)

    return onehot


DEFAULT_PATCHER.add(
    Patch(
        torch.nn.functional,
        onehot_wrapper(torch.nn.functional.one_hot),
        "one_hot",
    )
)


def noop_wrapper(fn):
    @wraps(fn)
    def noop(input: torch.Tensor, *args, **kwargs):
        return input

    return noop


DEFAULT_PATCHER.add(
    Patch(FakeTensor, noop_wrapper(FakeTensor.tolist), "tolist")
)

import warnings
_str = str
try:
    
    

    from torch.amp.autocast_mode import autocast, is_autocast_available

    # Hacky patch to get around the fact this init method has no handling for 'meta' tensors.
    def autoamp_init(
        self,
        device_type: str,
        dtype=None,
        enabled: bool = True,
        cache_enabled: Optional[bool] = None,
    ):
        if not isinstance(device_type, _str):
            raise ValueError(
                f"Expected `device_type` of type `str`, got: `{type(device_type)}`"
            )
        if dtype is None:
            if device_type == "meta":
                dtype = torch.get_autocast_dtype("cpu")
            else:
                dtype = torch.get_autocast_dtype(device_type)
        if torch._jit_internal.is_scripting():
            self._enabled = enabled
            self.device = device_type
            self.fast_dtype = dtype
            assert dtype is not None
            return
        self.device = device_type
        if not is_autocast_available(self.device):
            raise RuntimeError(
                f"User specified an unsupported autocast device_type '{self.device}'"
            )
        self.custom_backend_name = torch._C._get_privateuse1_backend_name()
        self.fast_dtype = torch.get_autocast_dtype(self.device)
        if self.device == self.custom_backend_name:
            necessary_funcs = [
                "get_amp_supported_dtype",
            ]
            message = f"Tried to use AMP with the `{self.custom_backend_name}` backend, but the backend has not "
            message += "registered a module or  the module miss some necessary funcs. The backend should register "
            message += "a module by `torch._register_device_module`, and the module must have these funcs: \n"
            message += "`get_amp_supported_dtype() -> List[torch.dtype]`. \n"

            assert hasattr(torch, self.custom_backend_name), message
            self.custom_device_mod = getattr(torch, self.custom_backend_name)
            for func in necessary_funcs:
                assert hasattr(self.custom_device_mod, func), (
                    message + f"But the func `{func}` is missing. \n"
                )

        self._cache_enabled = torch.is_autocast_cache_enabled()
        if (
            enabled
            and torch.cuda.amp.common.amp_definitely_not_available()
            and self.device == "cuda"
        ):
            warnings.warn(
                "User provided device_type of 'cuda', but CUDA is not available. Disabling"
            )
            enabled = False
        if dtype is not None:
            self.fast_dtype = dtype
        if cache_enabled is not None:
            self._cache_enabled = cache_enabled

        if self.device == "cpu":
            supported_dtype = [torch.bfloat16, torch.float16]
            if self.fast_dtype not in supported_dtype and enabled:
                error_message = "In CPU autocast, but the target dtype is not supported. Disabling autocast.\n"
                error_message += "CPU Autocast only supports dtype of "
                error_message += (
                    ", ".join(str(dtype) for dtype in supported_dtype)
                    + " currently."
                )
                warnings.warn(error_message)
                enabled = False
        elif self.device == "xpu":
            supported_dtype = [torch.bfloat16, torch.float16]
            if self.fast_dtype not in supported_dtype:
                error_message = "In XPU autocast, but the target dtype is not supported. Disabling autocast.\n"
                error_message += "XPU Autocast only supports dtypes of torch.bfloat16 and torch.float16 currently."
                warnings.warn(error_message)
                enabled = False
        elif self.device == "ipu":
            supported_dtypes = [torch.bfloat16, torch.float16]
            if self.fast_dtype not in supported_dtypes:
                error_message = "In IPU autocast, but the target dtype is not supported. Disabling autocast.\n"
                error_message += "IPU Autocast only supports dtypes of torch.bfloat16 and torch.float16 currently."
                warnings.warn(error_message)
                enabled = False
        elif self.device == "hpu":
            supported_dtype = [torch.bfloat16, torch.float16]
            if self.fast_dtype not in supported_dtype:
                error_message = "In HPU autocast, but the target dtype is not supported. Disabling autocast.\n"
                error_message += "HPU Autocast only supports dtypes of torch.bfloat16 and torch.float16 currently."
                warnings.warn(error_message)
                enabled = False
        elif self.device == self.custom_backend_name:
            supported_dtype = self.custom_device_mod.get_amp_supported_dtype()
            if self.fast_dtype not in supported_dtype:
                error_message = f"In {self.custom_backend_name} autocast, but the target dtype is not supported. "
                error_message += f"Disabling autocast.\n {self.custom_backend_name} Autocast only supports dtypes of "
                error_message += (
                    ", ".join(str(dtype) for dtype in supported_dtype)
                    + " currently."
                )
                warnings.warn(error_message)
                enabled = False
        elif self.device == "cuda":
            if (
                enabled
                and self.fast_dtype == torch.bfloat16
                and not torch.cuda.is_bf16_supported()
            ):
                raise RuntimeError(
                    "Current CUDA Device does not support bfloat16. Please switch dtype to float16."
                )
        elif self.device == "xla":
            supported_dtype = [torch.float16, torch.bfloat16]
            if self.fast_dtype not in supported_dtype:
                error_message = "In XLA autocast, but the target dtype is not supported. Disabling autocast.\n"
                error_message += "XLA Autocast only supports dtype of torch.bfloat16 currently."
                warnings.warn(error_message)
                enabled = False
        self._enabled = enabled

    from torch.amp.autocast_mode import autocast

    DEFAULT_PATCHER.add(Patch(autocast, autoamp_init, "__init__"))

except:
    pass

try:

    from accelerate.utils.modeling import (
        check_device_same,
        is_npu_available,
        is_xpu_available,
    )

    # Hacky patch to get around this function trying to set the parameter of a non meta tensor to meta.
    # Also handles FakeTensors.
    def set_module_tensor_to_device(
        module: torch.nn.Module,
        tensor_name: str,
        device: Union[int, str, torch.device],
        value: Optional[torch.Tensor] = None,
        dtype: Optional[Union[str, torch.dtype]] = None,
        fp16_statistics: Optional[torch.HalfTensor] = None,
        tied_params_map: Optional[
            Dict[int, Dict[torch.device, torch.Tensor]]
        ] = None,
    ):
        """
        A helper function to set a given tensor (parameter of buffer) of a module on a specific device (note that doing
        `param.to(device)` creates a new tensor not linked to the parameter, which is why we need this function).

        Args:
            module (`torch.nn.Module`):
                The module in which the tensor we want to move lives.
            tensor_name (`str`):
                The full name of the parameter/buffer.
            device (`int`, `str` or `torch.device`):
                The device on which to set the tensor.
            value (`torch.Tensor`, *optional*):
                The value of the tensor (useful when going from the meta device to any other device).
            dtype (`torch.dtype`, *optional*):
                If passed along the value of the parameter will be cast to this `dtype`. Otherwise, `value` will be cast to
                the dtype of the existing parameter in the model.
            fp16_statistics (`torch.HalfTensor`, *optional*):
                The list of fp16 statistics to set on the module, used for 8 bit model serialization.
            tied_params_map (Dict[int, Dict[torch.device, torch.Tensor]], *optional*, defaults to `None`):
                A map of current data pointers to dictionaries of devices to already dispatched tied weights. For a given
                execution device, this parameter is useful to reuse the first available pointer of a shared weight on the
                device for all others, instead of duplicating memory.
        """
        # Recurse if needed
        if "." in tensor_name:
            splits = tensor_name.split(".")
            for split in splits[:-1]:
                new_module = getattr(module, split)
                if new_module is None:
                    raise ValueError(f"{module} has no attribute {split}.")
                module = new_module
            tensor_name = splits[-1]

        if (
            tensor_name not in module._parameters
            and tensor_name not in module._buffers
        ):
            raise ValueError(
                f"{module} does not have a parameter or a buffer named {tensor_name}."
            )
        is_buffer = tensor_name in module._buffers
        old_value = getattr(module, tensor_name)

        # Treat the case where old_value (or a custom `value`, typically offloaded to RAM/disk) belongs to a tied group, and one of the weight
        # in the tied group has already been dispatched to the device, by avoiding reallocating memory on the device and just copying the pointer.
        if (
            value is not None
            and tied_params_map is not None
            and value.data_ptr() in tied_params_map
            and device in tied_params_map[value.data_ptr()]
        ):
            module._parameters[tensor_name] = tied_params_map[value.data_ptr()][
                device
            ]
            return
        elif (
            tied_params_map is not None
            and old_value.data_ptr() in tied_params_map
            and device in tied_params_map[old_value.data_ptr()]
        ):
            module._parameters[tensor_name] = tied_params_map[
                old_value.data_ptr()
            ][device]
            return

        if (
            old_value.device == torch.device("meta")
            and device not in ["meta", torch.device("meta")]
            and value is None
        ):
            raise ValueError(
                f"{tensor_name} is on the meta device, we need a `value` to put in on {device}."
            )

        if value is not None:
            if old_value.shape != value.shape:
                raise ValueError(
                    f'Trying to set a tensor of shape {value.shape} in "{tensor_name}" (which has shape {old_value.shape}), this look incorrect.'
                )

            if dtype is None:
                # For compatibility with PyTorch load_state_dict which converts state dict dtype to existing dtype in model
                value = value.to(old_value.dtype)
            elif not str(value.dtype).startswith(
                ("torch.uint", "torch.int", "torch.bool")
            ):
                value = value.to(dtype)

        param = (
            module._parameters[tensor_name]
            if tensor_name in module._parameters
            else None
        )
        param_cls = type(param)

        device_quantization = None
        with torch.no_grad():
            # leave it on cpu first before moving them to cuda
            # # fix the case where the device is meta, we don't want to put it on cpu because there is no data =0
            if (
                param is not None
                and param.device.type != "cuda"
                and torch.device(device).type == "cuda"
                and param_cls.__name__
                in ["Int8Params", "FP4Params", "Params4bit"]
            ):
                device_quantization = device
                device = "cpu"
            # `torch.Tensor.to(<int num>)` is not supported by `torch_npu` (see this [issue](https://github.com/Ascend/pytorch/issues/16)).
            if is_npu_available() and isinstance(device, int):
                device = f"npu:{device}"
            if is_xpu_available() and isinstance(device, int):
                device = f"xpu:{device}"
            if value is None:
                new_value = old_value.to(device)
                if dtype is not None and device in [
                    "meta",
                    torch.device("meta"),
                ]:
                    if not str(old_value.dtype).startswith(
                        ("torch.uint", "torch.int", "torch.bool")
                    ):
                        new_value = new_value.to(dtype)

                    if not is_buffer:
                        module._parameters[tensor_name] = param_cls(
                            new_value, requires_grad=old_value.requires_grad
                        )
            elif isinstance(value, torch.Tensor):
                new_value = value.to(device)
            else:
                new_value = torch.tensor(value, device=device)
            if device_quantization is not None:
                device = device_quantization
            if is_buffer:
                module._buffers[tensor_name] = new_value
            elif value is not None or not check_device_same(
                torch.device(device), module._parameters[tensor_name].device
            ):
                param_cls = type(module._parameters[tensor_name])
                kwargs = module._parameters[tensor_name].__dict__
                if param_cls.__name__ in ["Int8Params", "FP4Params"]:
                    if (
                        param_cls.__name__ == "Int8Params"
                        and new_value.dtype == torch.float32
                    ):
                        # downcast to fp16 if any - needed for 8bit serialization
                        new_value = new_value.to(torch.float16)
                    # quantize module that are going to stay on the cpu so that we offload quantized weights
                    if device == "cpu" and param_cls.__name__ == "Int8Params":
                        new_value = (
                            param_cls(
                                new_value,
                                requires_grad=old_value.requires_grad,
                                **kwargs,
                            )
                            .to(0)
                            .to("cpu")
                        )
                        new_value.CB = new_value.CB.to("cpu")
                        new_value.SCB = new_value.SCB.to("cpu")
                    else:
                        new_value = param_cls(
                            new_value,
                            requires_grad=old_value.requires_grad,
                            **kwargs,
                        ).to(device)
                elif param_cls.__name__ in ["QTensor", "QBitsTensor"]:
                    new_value = torch.nn.Parameter(
                        new_value, requires_grad=old_value.requires_grad
                    ).to(device)
                elif isinstance(new_value, FakeTensor) or isinstance(
                    old_value, FakeTensor
                ):
                    new_value = torch.nn.Parameter(
                        new_value, requires_grad=old_value.requires_grad
                    ).to(device)
                else:
                    new_value = param_cls(
                        new_value, requires_grad=old_value.requires_grad
                    ).to(device)

                module._parameters[tensor_name] = new_value
                if fp16_statistics is not None:
                    module._parameters[tensor_name].SCB = fp16_statistics.to(
                        device
                    )
                    del fp16_statistics
                # as we put the weight to meta, it doesn't have SCB attr anymore. make sure that it is not a meta weight
                if (
                    module.__class__.__name__ == "Linear8bitLt"
                    and getattr(module.weight, "SCB", None) is None
                    and str(module.weight.device) != "meta"
                ):
                    # quantize only if necessary
                    device_index = (
                        torch.device(device).index
                        if torch.device(device).type == "cuda"
                        else None
                    )
                    if (
                        not getattr(module.weight, "SCB", None)
                        and device_index is not None
                    ):
                        if (
                            module.bias is not None
                            and module.bias.device.type != "meta"
                        ):
                            # if a bias exists, we need to wait until the bias is set on the correct device
                            module = module.cuda(device_index)
                        elif module.bias is None:
                            # if no bias exists, we can quantize right away
                            module = module.cuda(device_index)
                elif (
                    module.__class__.__name__ == "Linear4bit"
                    and getattr(module.weight, "quant_state", None) is None
                ):
                    # quantize only if necessary
                    device_index = (
                        torch.device(device).index
                        if torch.device(device).type == "cuda"
                        else None
                    )
                    if (
                        not getattr(module.weight, "quant_state", None)
                        and device_index is not None
                    ):
                        module.weight = module.weight.cuda(device_index)
        # clean pre and post foward hook
        if is_npu_available():
            torch.npu.empty_cache()
        elif is_xpu_available():
            torch.xpu.empty_cache()
        else:
            torch.cuda.empty_cache()

        # When handling tied weights, we update tied_params_map to keep track of the tied weights that have already been allocated on the device in
        # order to avoid duplicating memory, see above.
        if (
            tied_params_map is not None
            and old_value.data_ptr() in tied_params_map
            and device not in tied_params_map[old_value.data_ptr()]
        ):
            tied_params_map[old_value.data_ptr()][device] = new_value
        elif (
            value is not None
            and tied_params_map is not None
            and value.data_ptr() in tied_params_map
            and device not in tied_params_map[value.data_ptr()]
        ):
            tied_params_map[value.data_ptr()][device] = new_value

    from accelerate import hooks

    DEFAULT_PATCHER.add(
        Patch(hooks, set_module_tensor_to_device, "set_module_tensor_to_device")
    )

except:
    pass


DEFAULT_PATCHER.__enter__()

from .contexts.GraphBasedContext import GlobalTracingContext

bool = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.bool
bytes = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.bytes
int = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.int
float = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.float
str = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.str
complex = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.complex
bytearray = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.bytearray
tuple = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.tuple
list = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.list
set = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.set
dict = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.dict
apply = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.apply
log = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.log
cond = GlobalTracingContext.GLOBAL_TRACING_CONTEXT.cond



from typing import Callable

from torch._subclasses.fake_tensor import FakeCopyMode, FakeTensorMode
from torch.fx.experimental.symbolic_shapes import ShapeEnv

from .Node import Node


def validate(target: Callable, *args, **kwargs):

    from ..contexts.GraphBasedContext import GlobalTracingContext

    # Enter FakeMode.
    with FakeTensorMode(
        allow_non_fake_inputs=True,
        shape_env=ShapeEnv(assume_static_by_default=True),
    ) as fake_mode:
        with FakeCopyMode(fake_mode):

            with GlobalTracingContext.exit_global_tracing_context():

                args, kwargs = Node.prepare_inputs((args, kwargs), proxy=True)

                return target(
                    *args,
                    **kwargs,
                )
